#!/usr/bin/env python
from pathlib import Path
from copy import deepcopy
import logging
import sys

from fiona.drvsupport import supported_drivers
import geopandas as gpd
import numpy as np
from shapely.geometry import MultiPolygon, Polygon
from shapely.ops import transform
import jigsawpy
from pyproj import Transformer

from ocsmesh import Raster, Geom, Hfun, Mesh
from ocsmesh import utils

logging.basicConfig(
    stream=sys.stdout,
    format='%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S'
    )
#logging.getLogger().setLevel(logging.DEBUG)
#logging.getLogger().setLevel(logging.INFO)

_logger = logging.getLogger(__name__)

# Enable KML driver
#from stackoverflow issue 72960340
supported_drivers['KML'] = 'rw'
supported_drivers['LIBKML'] = 'rw'

class RemeshByShape:

    @property
    def script_name(self):
        return 'remesh_by_shape'

    def __init__(self, sub_parser):

        this_parser = sub_parser.add_parser(
            self.script_name,
            help="Locally refine an existing input mesh based"
                 + " on input DEMs within the region specified"
                 + " by input shape and cut-off depth.")

        this_parser.add_argument('--in-crs', default='EPSG:4326')
        this_parser.add_argument('--shape', type=str)

        this_parser.add_argument('-u', '--upstream', action='store_true')
        this_parser.add_argument(
            '--cutoff', default=-250, type=float,
            help="Refinement cutoff depth in meters (positive up)")
        this_parser.add_argument('--factor', default=2, type=float)
        this_parser.add_argument(
            '--contour',
            action='append', nargs='+', type=float, dest='contours',
            metavar='CONTOUR_DEFN', default=[],
            help="Each contour's (level, [expansion, target])"
                 " to be applied on all size functions in collector")
        this_parser.add_argument(
            '--patch',
            action='append', nargs=3, type=float, dest='patches',
            metavar='PATCH_DEFN', default=[],
            help="Specify patch mesh size above a given contour level"
                 " by passing (lower_bound, expansion, target_size)"
                 " for each patch")
        this_parser.add_argument(
            '--constant',
            action='append', nargs=2, type=float, dest='constants',
            metavar='CONST_DEFN', default=[],
            help="Specify constant mesh size above a given contour level"
                 " by passing (lower_bound, target_size) for each constant")
        this_parser.add_argument('-s', '--sieve', type=float)

        this_parser.add_argument(
            '--interpolate', nargs='+', type=Path, default=[],
            help="To interpolate from depth of DEMs not involved in"
                 " the remeshing process")

        this_parser.add_argument('-o', '--output', type=Path)
        this_parser.add_argument('-f', '--output-format', default='2dm')
        this_parser.add_argument('--nprocs', type=int, default=-1)

        this_parser.add_argument('mesh', type=Path)


    def run(self, args):

        # Get inputs
        mesh_file = args.mesh
        mesh_crs = args.mesh_crs

        shape_path = args.shape

        refine_upstream = args.upstream
        refine_factor = args.factor
        refine_cutoff = args.cutoff

        contours = args.contours
        patches = args.patches
        constants = args.constants

        sieve = args.sieve
        interp = args.interpolate

        out_path = args.output
        out_format = args.output_format
        nprocs = args.nprocs

        # Process inputs
        contour_defns = []
        for contour in contours:
            if len(contour) > 3:
                raise ValueError(
                    "Invalid format for contour specification."
                    " It should be level [expansion target-size].")
            level, expansion_rate, target_size = [
                    *contour, *[None]*(3-len(contour))]
            contour_defns.append((level, expansion_rate, target_size))

        patch_defns = []
        for lower_bound, target_size in patches:
            patch_defns.append((lower_bound, expansion_rate, target_size))

        constant_defns = []
        for lower_bound, target_size in constants:
            constant_defns.append((lower_bound, target_size))


        interp_rast_list = []
        for dem in interp:
            interp_rast_list.append(Raster(dem))

        mesh = Mesh.open(mesh_file, crs=mesh_crs)

        geom = Geom(deepcopy(mesh))
        geom_jig = geom.msh_t()

        initial_hfun = Hfun(deepcopy(mesh))
        initial_hfun.size_from_mesh()

        # DO NOT DEEPCOPY
        initial_hfun_jig = initial_hfun.msh_t()
        ref_crs = initial_hfun_jig.crs

        utils.reproject(geom_jig, ref_crs)

        init_jig = None

        # If there's an input shape, refine in the shape, otherwise
        # refine the whole domain by the factor
        if shape_path or refine_cutoff is not None:

            mesh_poly = mesh.hull.multipolygon()
            gdf_mesh_poly = gpd.GeoDataFrame(
                    geometry=gpd.GeoSeries(mesh_poly), crs=mesh.crs)

            if shape_path:
                gdf_shape = gpd.read_file(shape_path)

                gdf_shape = gdf_shape.to_crs(mesh.crs)


                gdf_to_refine = gpd.overlay(
                        gdf_mesh_poly, gdf_shape, how='intersection')

                gdf_diff = gpd.overlay(
                        gdf_mesh_poly, gdf_shape, how='difference')
                diff_polys = []
                for geom in gdf_diff.geometry:
                    if isinstance(geom, Polygon):
                        diff_polys.append(geom)
                    elif isinstance(geom, MultiPolygon):
                        diff_polys.extend(geom)

                if refine_upstream:
                    # TODO: Check for multipolygon and single polygon in multi assumption
                    area_ref = 0.05 * np.sum(
                            [i.area for i in gdf_to_refine.geometry])
                    upstream_polys = []
                    for ipoly in diff_polys:
                        if ipoly.area < area_ref:
                            upstream_polys.append(ipoly)
                    if upstream_polys:
                        gdf_upstream = gpd.GeoDataFrame(
                                geometry=gpd.GeoSeries(upstream_polys),
                                crs=gdf_diff.crs)

                        gdf_to_refine = gpd.overlay(
                                gdf_upstream, gdf_to_refine, how='union')
            else:
                gdf_to_refine = gdf_mesh_poly

            gdf_to_refine = gdf_to_refine.to_crs(ref_crs)

            if refine_cutoff is not None:
                cutoff_mp = mesh.get_multipolygon(zmin=refine_cutoff)
                cutoff_gdf = gpd.GeoDataFrame(
                        geometry=gpd.GeoSeries(cutoff_mp), crs=mesh.crs)
                cutoff_gdf = cutoff_gdf.to_crs(ref_crs)
                gdf_to_refine = gpd.overlay(
                        gdf_to_refine, cutoff_gdf, how='intersection')

            refine_polys = gdf_to_refine.unary_union

            # Initial mesh for the refinement (all except refinement area)
            init_jig = deepcopy(mesh.msh_t)
            utils.reproject(init_jig, ref_crs)
            utils.clip_mesh_by_shape(
                    init_jig,
                    refine_polys,
                    fit_inside=True,
                    inverse=True,
                    in_place=True)

            # Fix elements in the inital mesh that are NOT clipped by refine
            # polygon
            init_jig.vert2['IDtag'][:] = -1

            # Preparing refinement size function
            vert_in = utils.get_verts_in_shape(initial_hfun_jig, refine_polys)
            # Reduce hfun by factor in refinement area; modifying in-place

            refine_hfun_jig = utils.clip_mesh_by_shape(
                initial_hfun_jig, refine_polys, fit_inside=False)

            utils.clip_mesh_by_shape(
                initial_hfun_jig, refine_polys,
                fit_inside=True, inverse=True, in_place=True)


        else:
            # Refine the whole domain by factor
            refine_hfun_jig = deepcopy(initial_hfun_jig)

        # Prepare refinement size function with additional criteria
        refine_hfun_jig.value[:] = refine_hfun_jig.value / refine_factor

        hfun_refine = Hfun(Mesh(deepcopy(refine_hfun_jig)))

        transformer = Transformer.from_crs(
                mesh.crs, ref_crs, always_xy=True)
        for level, expansion_rate, target_size in contour_defns:
            if expansion_rate is None:
                expansion_rate = 0.1
            if target_size is None:
                target_size = np.min(refine_hfun_jig.value)

            refine_ctr = mesh.get_contour(level=level)
            refine_ctr = transform(transformer.transform, refine_ctr)

            hfun_refine.add_feature(
                    refine_ctr, expansion_rate, target_size,
                    nprocs=nprocs)

        for lower_bound, expansion_rate, target_size in patch_defns:
            refine_mp = mesh.get_multipolygon(zmin=lower_bound)
            refine_mp = transform(transformer.transform, refine_mp)

            hfun_refine.add_patch(
                    refine_mp, expansion_rate, target_size, nprocs)

        for lower_bound, target_size in constant_defns:
            refine_mp = mesh.get_multipolygon(zmin=lower_bound)
            refine_mp = transform(transformer.transform, refine_mp)

            hfun_refine.add_patch(
                    refine_mp, None, target_size, nprocs)

        refine_hfun_jig = hfun_refine.msh_t()
        utils.reproject(refine_hfun_jig, ref_crs)

        final_hfun_jig = utils.merge_msh_t(
                initial_hfun_jig, refine_hfun_jig,
                out_crs=ref_crs,
                drop_by_bbox=False)

        if not (geom_jig.crs == ref_crs
                and (init_jig and init_jig.crs == ref_crs)):
            raise ValueError(
                "CRS for geometry, hfun and init mesh is not the same")

        opts = jigsawpy.jigsaw_jig_t()
        opts.hfun_scal = "absolute"
        opts.hfun_hmin = np.min(final_hfun_jig.value)
        opts.hfun_hmax = np.max(final_hfun_jig.value)
        opts.mesh_dims = +2

        remesh_jig = jigsawpy.jigsaw_msh_t()
        remesh_jig.mshID = 'euclidean-mesh'
        remesh_jig.ndims = 2
        remesh_jig.crs = init_jig.crs

        jigsawpy.lib.jigsaw(
                opts, geom_jig, remesh_jig,
                init=init_jig,
                hfun=final_hfun_jig)

        utils.finalize_mesh(remesh_jig, sieve)

        # Interpolate from inpu mesh and DEM if any
        utils.interpolate_euclidean_mesh_to_euclidean_mesh(
                mesh.msh_t, remesh_jig)
        final_mesh = Mesh(remesh_jig)
        if interp_rast_list:
            final_mesh.interpolate(interp_rast_list, nprocs=nprocs)

        # Write to disk
        final_mesh.write(
                str(out_path), format=out_format, overwrite=True)



#!/bin/env python3
import gc
import sys
import pathlib
import logging

import geopandas as gpd
from shapely.geometry import MultiPolygon

from ocsmesh import Raster, Geom, Hfun, JigsawDriver
from ocsmesh.mesh.mesh import Mesh
from ocsmesh.geom.shapely import MultiPolygonGeom
from ocsmesh.hfun.mesh import HfunMesh
from ocsmesh.features.contour import Contour
from ocsmesh.mesh.parsers import sms2dm
from ocsmesh.utils import msh_t_to_2dm


logging.basicConfig(
    stream=sys.stdout,
    format='%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S'
    )
#logging.getLogger().setLevel(logging.DEBUG)
#logging.getLogger().setLevel(logging.INFO)


class MeshUpgrader:

    @property
    def script_name(self):
        return 'mesh_upgrader'

    def __init__(self, sub_parser):
        # e.g
        # ./multi_dem_mesh_2.py \
        #       --basemesh data/prusvi/mesh/PRUSVI_COMT.14 \
        #       --demlo gebco_2020_n90.0_s0.0_w-90.0_e0.0.tif \
        #       --demhi ncei19_*.tif \
        #       --out demo/final_mesh.2dm

        this_parser = sub_parser.add_parser(
            self.script_name,
            help="Create a new mesh based on a base-mesh and input DEMs.")

        this_parser.add_argument('--basemesh', required=True)
        this_parser.add_argument('--demlo', nargs='*', required=True)
        this_parser.add_argument('--demhi', nargs='*', required=True)
        this_parser.add_argument('--out', required=True)

    def run(self, args):

        logging.info(args)

        base_path = pathlib.Path(args.basemesh)
        demlo_paths = args.demlo
        demhi_paths = args.demhi
        out_path = pathlib.Path(args.out)

        out_path.parent.mkdir(exist_ok=True, parents=True)

        base_mesh_4_hfun = Mesh.open(base_path, crs="EPSG:4326")
        base_mesh_4_geom = Mesh.open(base_path, crs="EPSG:4326")

        geom_rast_list = []
        hfun_rast_list = []
        hfun_hirast_list = []
        hfun_lorast_list = []
        interp_rast_list = []
        for dem_path in demlo_paths:
            hfun_lorast_list.append(Raster(dem_path))
            interp_rast_list.append(Raster(dem_path))

        for dem_path in demhi_paths:
            geom_rast_list.append(Raster(dem_path))
            hfun_hirast_list.append(Raster(dem_path))
            interp_rast_list.append(Raster(dem_path))


        hfun_rast_list = [*hfun_lorast_list, *hfun_hirast_list]

        geom = Geom(
            geom_rast_list, base_mesh=base_mesh_4_geom,
            zmax=15, nprocs=4)

        hfun = Hfun(
            hfun_rast_list, base_mesh=base_mesh_4_hfun,
            hmin=30, hmax=15000, nprocs=4)

        ## Add contour refinements at 0 separately for GEBCO and NCEI
        ctr1 = Contour(level=0, sources=hfun_hirast_list)
        hfun.add_contour(None, 1e-3, 30, contour_defn=ctr1)

        ctr2 = Contour(level=0, sources=hfun_lorast_list)
        hfun.add_contour(None, 1e-2, 500, contour_defn=ctr2)

        ## Add constant values from 0 to inf on hi-res rasters
        hfun.add_constant_value(30, 0, source_index=list(range(len(demhi_paths))))


        # Calculate geom
        geom_mp = geom.get_multipolygon()
        # Write to disk
        gpd.GeoDataFrame(
                {'geometry': geom_mp},
                crs="EPSG:4326"
                ).to_file(str(out_path) + '.geom.shp')
        del geom_mp

        # Calculate hfun
        hfun_msh_t = hfun.msh_t()
        # Write to disk
        sms2dm.writer(
                msh_t_to_2dm(hfun_msh_t),
                str(out_path) + '.hfun.2dm',
                True)
        del hfun_msh_t


        # Read back stored values to pass to mesh driver
        read_gdf = gpd.read_file(str(out_path) + '.geom.shp')
        geom_from_disk = MultiPolygonGeom(
            MultiPolygon(list(read_gdf.geometry)),
            crs=read_gdf.crs)

        read_hfun = Mesh.open(str(out_path) + '.hfun.2dm', crs="EPSG:4326")
        hfun_from_disk = HfunMesh(read_hfun)

        jigsaw = JigsawDriver(geom_from_disk, hfun=hfun_from_disk, initial_mesh=None)
        jigsaw.verbosity = 1

        ## Execute mesher (processing of geom and hfun happens here)
        mesh = jigsaw.run()

        ## Free-up memory
        del read_gdf
        del geom_from_disk
        del read_hfun
        del hfun_from_disk
        gc.collect()

        mesh.write(str(out_path) + '.raw.2dm', format='2dm', overwrite=True)

        ## Interpolate DEMs on the mesh
        mesh.interpolate(interp_rast_list, nprocs=4)

        ## Output
        mesh.write(out_path, format='2dm', overwrite=True)



#! python
import unittest
import tempfile
from pathlib import Path

import geopandas as gpd
from pyproj import CRS
from shapely import geometry

from ocsmesh.features.linefeature import LineFeature

class LineFeatureCapabilities(unittest.TestCase):

    def setUp(self):
        self._crs_1 = CRS.from_epsg(4326)
        self._crs_2 = CRS.from_string('nad83')
        self._lines_1 = geometry.LineString(
            [ 
                [0, 0], [1, 1], [2, 2]
            ]
        )
        self._lines_2 = geometry.LineString(
            [
                [1, 0], [2, 1]
            ]
        )
        self._mult_line_1 = geometry.MultiLineString(
            [self._lines_1, self._lines_2]
        )


    def test_ctor_args_1(self):
        lf = LineFeature(self._lines_1)
        shape, crs = lf.get_multiline()
        self.assertIsInstance(shape, geometry.MultiLineString)
        self.assertEqual(crs, self._crs_1)


    def test_ctor_args_2(self):
        lf = LineFeature(self._mult_line_1, self._crs_2)
        shape, crs = lf.get_multiline()
        self.assertIsInstance(shape, geometry.MultiLineString)
        self.assertEqual(crs, self._crs_2)


    def test_ctor_args_3(self):
        lf = LineFeature(shape=self._lines_1, shape_crs=self._crs_2)
        shape, crs = lf.get_multiline()
        self.assertIsInstance(shape, geometry.MultiLineString)
        self.assertEqual(crs, self._crs_2)


    def test_ctor_args_4(self):

        with tempfile.TemporaryDirectory() as tdir:

            shape_path = tdir

            gdf = gpd.GeoDataFrame(geometry=[self._mult_line_1], crs=self._crs_2)
            gdf.to_file(shape_path)

            lf = LineFeature(shapefile=shape_path)
            shape, crs = lf.get_multiline()
            self.assertIsInstance(shape, geometry.MultiLineString)
            self.assertEqual(crs, self._crs_2)


    def test_ctor_args_5(self):

        with tempfile.TemporaryDirectory() as tdir:

            # NOTE: Path is file this time
            shape_path = Path(tdir) / 'test_shape.shp'

            gdf = gpd.GeoDataFrame(geometry=[self._mult_line_1], crs=self._crs_2)
            gdf.to_file(shape_path)

            lf = LineFeature(shapefile=shape_path)
            shape, crs = lf.get_multiline()
            self.assertIsInstance(shape, geometry.MultiLineString)
            self.assertEqual(crs, self._crs_2)


    def test_ctor_args_6(self):

        with tempfile.TemporaryDirectory() as tdir:

            shape_path = tdir

            # NOTE: This time shapefile doesn't have CRS
            gdf = gpd.GeoDataFrame(geometry=[self._mult_line_1])
            gdf.to_file(shape_path)

            lf = LineFeature(shapefile=shape_path)
            with self.assertRaises(ValueError):
                lf.get_multiline()


if __name__ == '__main__':
    unittest.main()



"""This module defines mesh based geometry class
"""

import os
from typing import Union

# from jigsawpy import jigsaw_msh_t  # type: ignore[import]
# import matplotlib.pyplot as plt  # type: ignore[import]
# import mpl_toolkits.mplot3d as m3d  # type: ignore[import]
# import numpy as np  # type: ignore[import]
from shapely.geometry import MultiPolygon

from ocsmesh.geom.base import BaseGeom
from ocsmesh.mesh.mesh import Mesh
from ocsmesh.mesh.base import BaseMesh


class MeshDescriptor:
    """Descriptor class for storing handle to input mesh"""

    def __set__(self, obj, val: Union[BaseMesh, str, os.PathLike]):

        if isinstance(val, (str, os.PathLike)):  # type: ignore[misc]
            val = Mesh.open(val)

        if not isinstance(val, BaseMesh):
            raise TypeError(f'Argument mesh must be of type {Mesh}, {str} '
                            f'or {os.PathLike}, not type {type(val)}')

        obj.__dict__['mesh'] = val

    def __get__(self, obj, objtype=None) -> BaseMesh:
        return obj.__dict__['mesh']


class MeshGeom(BaseGeom):
    """Mesh based geometry.

    Create a geometry based on an input mesh. All the calculations
    for polygon are done on the underlying mesh object.

    Attributes
    ----------
    mesh : BaseMesh
        Reference to the underlying mesh object used to create
        the geometry
    crs : CRS
        CRS of the underlying mesh

    Methods
    -------
    get_multipolygon(**kwargs)
        Returns `shapely` object representation of the geometry
    msh_t(**kwargs)
        Returns the `jigsawpy` vertex-edge representation of the geometry

    Notes
    -----
    This class is a handy tool for reusing existing mesh extent
    for geometry (i.e. domain) definition and then remeshing that
    domain.
    """

    _mesh = MeshDescriptor()

    def __init__(
            self,
            mesh: Union[BaseMesh, str, os.PathLike]
            ) -> None:
        """ Initialize a mesh based geometry object

        Parameters
        ----------
        mesh:
            Input object used to compute the output mesh hull.
        """

        self._mesh = mesh

    def get_multipolygon(self, **kwargs) -> MultiPolygon:
        """Returns the `shapely` representation of the geometry

        Calculates and returns the `MultiPolygon` representation of
        the geometry.

        Parameters
        ----------
        **kwargs : dict, optional
            Currently unused for this class, needed for generic API
            support

        Returns
        -------
        MultiPolygon
            Calculated polygon from mesh based on the element boundaries
        """

        # TODO: What if there's no tria, e.g. Mesh object is
        # created from geom.msh_t() return value
        return self.mesh.hull.multipolygon()

    @property
    def mesh(self):
        """Read-only attribute for reference to the source mesh"""
        return self._mesh

    @property
    def crs(self):
        """Read-only attribute returning the CRS of the underlying mesh"""
        return self._mesh.crs



#! python
import tempfile
import unittest
import warnings
from pathlib import Path

import numpy as np
from jigsawpy import jigsaw_msh_t
from shapely import geometry

from ocsmesh.mesh.mesh import Mesh


def edge_at (x, y):
    return geometry.Point(x, y).buffer(0.05)

class BoundaryExtraction(unittest.TestCase):

    def setUp(self):
        """
        Note:
            x = x-index
            y = y-index

            node-index(node-id)

              25(26)             29(30)
          5     *---*---*---*---*
                | \ | \ | \ | \ |
          4     *---*---*---*---*
                | \ | \ | \ | \ |
          3     *---*---*---*---*
                | \ |   | \ | \ |
          2     *---*---*---*---*
                | \ | \ | \ | \ |
          1     *---*---*---*---*
                | \ | \ | \ | \ |
          0     *---*---*---*---*
              0(1)               4(5)

                0   1   2   3   4
        """

        # Create a basic grid with a hole
        nx, ny = 5, 6
        hole_square = 10
        X, Y = np.meshgrid(range(nx), range(ny))
        verts = np.array(list(zip(X.ravel(), Y.ravel())))
        cells = []
        for j in range(ny - 1):
            for i in range(nx - 1):
                if (i + 1) + ((nx-1) * j) == hole_square:
                    continue
                cells.append([j * nx + i, j * nx + (i + 1), (j + 1) * nx + i])
                cells.append([j * nx + (i + 1), (j + 1) * nx + (i + 1), (j + 1) * nx + i])
        # NOTE: Everywhere is above 0 (auto: land) unless modified later
        vals = np.ones((len(verts), 1)) * 10

        # TODO: Replace with "make_mesh" util function
        mesh_msht = jigsaw_msh_t()
        mesh_msht.ndims = +2
        mesh_msht.mshID = 'euclidean-mesh'
        mesh_msht.tria3 = np.array(
            [(c, 0) for c in cells], dtype=jigsaw_msh_t.TRIA3_t
        )
        mesh_msht.vert2 = np.array(
            [(v, 0) for v in verts], dtype=jigsaw_msh_t.VERT2_t
        )
        mesh_msht.value = np.array(
            vals, dtype=jigsaw_msh_t.REALS_t
        )

        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', category=UserWarning, message='Input mesh has no CRS information'
            )

            self.mesh = Mesh(mesh_msht)

    def test_auto_boundary_fails_if_na_elev(self):
        # Set one node to nan value
        self.mesh.msh_t.value[-1] = np.nan
        with self.assertRaises(ValueError):
            self.mesh.boundaries.auto_generate()


    def test_auto_boundary_1open_correctness(self):
        # Set right boundary to be open
        self.mesh.msh_t.value[self.mesh.msh_t.vert2['coord'][:, 0] > 3] = -10

        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        # Mesh has one segment of each boundary type
        self.assertEqual(len(bdry.open()), 1)
        self.assertEqual(len(bdry.land()), 1)
        self.assertEqual(len(bdry.interior()), 1)

        # Boundaries node ID list
        self.assertEqual(bdry.open().iloc[0]['index_id'], [30, 25, 20, 15, 10, 5])
        self.assertEqual(
            bdry.land().iloc[0]['index_id'],
            [5, 4, 3, 2, 1, 6, 11, 16, 21, 26, 27, 28, 29, 30]
        )
        self.assertEqual(bdry.interior().iloc[0]['index_id'], [12, 13, 18, 17, 12])


    def test_auto_boundary_2open_correctness(self):
        # Set left and right boundary to be open
        self.mesh.msh_t.value[self.mesh.msh_t.vert2['coord'][:, 0] > 3] = -10
        self.mesh.msh_t.value[self.mesh.msh_t.vert2['coord'][:, 0] < 1] = -10

        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        self.assertEqual(len(bdry.open()), 2)
        self.assertEqual(len(bdry.land()), 2)
        self.assertEqual(len(bdry.interior()), 1)

        # Boundaries node ID list
        self.assertEqual(bdry.open().iloc[0]['index_id'], [1, 6, 11, 16, 21, 26])
        self.assertEqual(bdry.open().iloc[1]['index_id'], [30, 25, 20, 15, 10, 5])
        self.assertEqual(bdry.land().iloc[0]['index_id'], [5, 4, 3, 2, 1])
        self.assertEqual(bdry.land().iloc[1]['index_id'], [26, 27, 28, 29, 30])
        self.assertEqual(bdry.interior().iloc[0]['index_id'], [12, 13, 18, 17, 12])


    def test_manual_boundary_specification_correctness(self):
        # Shape for wrapping bottom boundary
        shape1 = geometry.box(0.5, -0.5, 3.5, 0.5)
        shape2 = geometry.box(1.5, -0.5, 2.5, 0.5)

        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        bdry.set_open(region=shape1)
        bdry.set_land(region=shape2)

        self.assertEqual(len(bdry.open()), 2)
        self.assertEqual(len(bdry.land()), 2)
        self.assertEqual(len(bdry.interior()), 1)

        # Boundaries node ID list
        self.assertEqual(bdry.open().iloc[0]['index_id'], [1, 2])
        self.assertEqual(bdry.open().iloc[1]['index_id'], [4, 5])
        self.assertEqual(
            bdry.land().iloc[0]['index_id'],
            [1, 6, 11, 16, 21, 26, 27, 28, 29, 30, 25, 20, 15, 10, 5]
        )
        self.assertEqual(bdry.land().iloc[1]['index_id'], [2, 3, 4])
        self.assertEqual(bdry.interior().iloc[0]['index_id'], [12, 13, 18, 17, 12])


    def test_manual_boundary_notaffect_interior(self):
        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        bdry.set_open(region=geometry.box(0.5, 1.5, 2.5, 3.5))

        self.assertEqual(len(bdry.open()), 0)
        self.assertEqual(len(bdry.land()), 1)
        self.assertEqual(len(bdry.interior()), 1)


    def test_manual_boundary_convex_region(self):
        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        bdry.set_open(region=geometry.Polygon([
            (-1, 4.5),
            (0.5, 4.5),
            (0.5, 3.5),
            (-0.5, 3.5),
            (-0.5, 1.5),
            (0.5, 1.5),
            (0.5, 0.5),
            (-1, 0.5),
        ]))

        self.assertEqual(len(bdry.open()), 2)
        self.assertEqual(len(bdry.land()), 2)
        self.assertEqual(len(bdry.interior()), 1)

        self.assertEqual(bdry.open().iloc[0]['index_id'], [1, 6, 11])
        self.assertEqual(bdry.open().iloc[1]['index_id'], [16, 21, 26])
        self.assertEqual(bdry.land().iloc[0]['index_id'], [11, 16])
        self.assertEqual(
            bdry.land().iloc[1]['index_id'],
            [26, 27, 28, 29, 30, 25, 20, 15, 10, 5, 4, 3, 2, 1]
        )


    def test_specified_boundary_order_nomerge(self):
        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        bdry.set_open(region=edge_at(1, 0))
        bdry.set_open(region=edge_at(4, 1))
        bdry.set_open(region=edge_at(0, 5))
        bdry.set_open(region=edge_at(4, 5))

        self.assertEqual(len(bdry.open()), 4)
        self.assertEqual(len(bdry.land()), 4)
        self.assertEqual(len(bdry.interior()), 1)

        self.assertEqual(bdry.open().iloc[0]['index_id'], [1, 2, 3])
        self.assertEqual(bdry.open().iloc[1]['index_id'], [5, 10, 15])
        self.assertEqual(bdry.open().iloc[2]['index_id'], [21, 26, 27])
        self.assertEqual(bdry.open().iloc[3]['index_id'], [29, 30, 25])


    def test_manual_boundary_brokenring_stillconnected(self):
        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        bdry.set_open(region=edge_at(4, 5))

        # Mesh has one segment of each boundary type
        self.assertEqual(len(bdry.open()), 1)
        self.assertEqual(len(bdry.land()), 1)
        self.assertEqual(len(bdry.interior()), 1)


    def test_manual_boundary_merge_sametype(self):
        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        bdry.set_land(region=edge_at(1, 0), merge=True)
        bdry.set_open(region=edge_at(4, 3), merge=True)
        bdry.set_open(region=edge_at(4, 4), merge=True)


        # Mesh has one segment of each boundary type
        self.assertEqual(len(bdry.open()), 1)
        self.assertEqual(len(bdry.land()), 1)
        self.assertEqual(len(bdry.interior()), 1)


    def test_manual_boundary_connect_two_separate_segments(self):
        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        bdry.set_open(region=edge_at(0, 0), merge=True)
        bdry.set_open(region=edge_at(4, 0), merge=True)

        self.assertEqual(len(bdry.open()), 2)
        self.assertEqual(len(bdry.land()), 2)
        self.assertEqual(len(bdry.interior()), 1)

        self.mesh.boundaries.set_open(region=edge_at(2, 0), merge=True)

        self.assertEqual(len(bdry.open()), 1)
        self.assertEqual(len(bdry.land()), 1)
        self.assertEqual(len(bdry.interior()), 1)


    def test_specified_boundary_order_withmerge(self):
        self.mesh.boundaries.auto_generate()
        # bdry is referring to mesh object and can be mutated
        bdry = self.mesh.boundaries

        bdry.set_open(region=edge_at(1, 0))
        bdry.set_open(region=edge_at(4, 1))
        bdry.set_open(region=edge_at(0, 5))
        bdry.set_open(region=edge_at(4, 5))
        bdry.set_open(region=edge_at(4, 4))
        bdry.set_open(region=edge_at(0, 0), merge=True)
        bdry.set_open(region=edge_at(0, 4), merge=True)

        self.assertEqual(len(bdry.open()), 4)
        self.assertEqual(len(bdry.land()), 4)
        self.assertEqual(len(bdry.interior()), 1)

        self.assertEqual(bdry.open().iloc[0]['index_id'], [6, 1, 2, 3])
        self.assertEqual(bdry.open().iloc[1]['index_id'], [5, 10, 15])
        self.assertEqual(bdry.open().iloc[2]['index_id'], [16, 21, 26, 27])
        self.assertEqual(bdry.open().iloc[3]['index_id'], [20, 25, 30, 29])


    def test_specify_boundary_on_imported_mesh_with_boundary(self):
        self.mesh.boundaries.auto_generate()

        with tempfile.NamedTemporaryFile(suffix='.grd') as fo:
            self.mesh.write(fo.name, format='grd', overwrite=True)
            imported_mesh = Mesh.open(fo.name)

        bdry = imported_mesh.boundaries

        bdry.set_open(region=edge_at(1, 0))

        self.assertEqual(len(bdry.open()), 1)
        self.assertEqual(len(bdry.land()), 1)
        self.assertEqual(len(bdry.interior()), 1)

        self.assertEqual(bdry.open().iloc[0]['index_id'], [1, 2, 3])


    def test_auto_find_islands_only(self):
        bdry = self.mesh.boundaries

        self.assertEqual(len(bdry.interior()), 0)

        bdry.find_islands()

        self.assertEqual(len(bdry.interior()), 1)
        self.assertEqual(bdry.interior().iloc[0]['index_id'], [12, 13, 18, 17, 12])


    def test_specify_boundary_on_mesh_with_no_boundary(self):
        bdry = self.mesh.boundaries

        with self.assertWarns(UserWarning) as w:
            bdry.set_open(region=edge_at(1, 0))
        self.assertTrue('didn\'t have prior boundary' in str(w.warning))

        self.assertEqual(len(bdry.open()), 1)
        self.assertEqual(len(bdry.land()), 0)
        self.assertEqual(len(bdry.interior()), 0)

        self.assertEqual(bdry.open().iloc[0]['index_id'], [1, 2, 3])



"""This module defines the base class for all size function types
"""

from abc import ABC, abstractmethod

from jigsawpy import jigsaw_msh_t  # type: ignore[import]


class BaseHfun(ABC):
    """Abstract base class used to construct OCSMesh "hfun" objects.

    This abstract class defines the interface that all size function
    objects are expected to have in OCSMesh. All size function
    subclasses need to inherit this base class.

    Methods
    -------
    msh_t()
        Returns the `jigsawpy` vertex-edge representation of the mesh.
    """


    @abstractmethod
    def msh_t(self) -> jigsaw_msh_t:
        '''Abstract method to generate hfun object.'''



from ocsmesh.geom.geom import Geom
from ocsmesh.geom.raster import RasterGeom
from ocsmesh.geom.mesh import MeshGeom
from ocsmesh.geom.shapely import PolygonGeom, MultiPolygonGeom
from ocsmesh.geom.collector import GeomCollector
__all__ = [
    "Geom",
    "RasterGeom",
    "MeshGeom",
    "PolygonGeom",
    "MultiPolygonGeom",
    "GeomCollector"
]



from abc import ABC, abstractmethod

class ContourBase(ABC):

    def __init__(self, sources=None):
        if sources is None:
            sources = []

        self._sources = []
        if not isinstance(sources, (list, tuple)):
            sources = [sources]
        for source in sources:
            self.add_source(source)

    @property
    def has_source(self):
        return len(self._sources) > 0

    @property
    def sources(self):
        return self._sources

    def add_source(self, source_object):
        src_class = type(source_object).__name__
        if src_class not in ("Raster", "RasterGeom", "HfunRaster"):
            # TODO: Accepts meshhfun?
            raise TypeError("")

        self._sources.append(source_object)

    def iter_contours(self):
        for source in self._sources:
            yield self._get_contour_from_source(source)

    @abstractmethod
    def _get_contour_from_source(self, source):
        pass


    @property
    @abstractmethod
    def level(self):
        return 0


class Contour(ContourBase):

    def __init__(self, level=None, sources=None):

        if sources is None:
            sources = []

        super().__init__(sources)
        self._level = level

    def _get_contour_from_source(self, source):
        src_class = type(source).__name__
        if src_class == "Raster":
            contour = source.get_contour(self._level)
            crs = source.crs
        elif src_class in ("RasterGeom", "HfunRaster"):
            contour = source.raster.get_contour(self._level)
            crs = source.raster.crs
        else:
            raise TypeError("")

        return contour, crs

    @property
    def level(self):
        return self._level


class FilledContour(ContourBase):

    def __init__(self,
                 level0=None,
                 level1=None,
                 sources=None,
                 max_contour_defn : Contour = None,
                 ):

        if sources is None:
            sources = []

        super().__init__(sources)
        if max_contour_defn:
            self._level0 = None
            self._level1 = max_contour_defn.level
            # NOTE: Overriding the existing sources
            self._sources = max_contour_defn.sources
        else:
            self._level0 = level0
            self._level1 = level1

    def _get_contour_from_source(self, source):
        z_info = {}
        if self._level0 is not None:
            z_info['zmin'] = self._level0
        if self._level1 is not None:
            z_info['zmax'] = self._level1

        src_class = type(source).__name__
        if src_class == "Raster":
            contour = source.get_multipolygon(**z_info)
            crs = source.crs
        elif src_class in ("RasterGeom", "HfunRaster"):
            contour = source.raster.get_multipolygon(**z_info)
            crs = source.raster.crs
        else:
            raise TypeError("")

        return contour, crs

    @property
    def level(self):
        return self._level0, self._level1



"""This module defines `shapely` object based geometry
"""

from typing import Union, Any

from pyproj import CRS
from shapely.geometry import Polygon, MultiPolygon

from ocsmesh.geom.base import BaseGeom


class ShapelyGeom(BaseGeom):
    """Base class for geoms based on shapely objects"""


class PolygonGeom(ShapelyGeom):
    """Geometry based on `shapely.geometry.Polygon`.

    Attributes
    ----------
    polygon : Polygon
        Reference to the underlying `shapely` object.
    crs : crs
        CRS associated with the geometry during construction.

    Methods
    -------
    get_multipolygon(**kwargs)
        Returns `shapely` object representation of the geometry
    msh_t(**kwargs)
        Returns the `jigsawpy` vertex-edge representation of the geometry
    """

    def __init__(self, polygon: Polygon, crs: Union[CRS, str]) -> None:
        """Initialize a `shapely` based geometry object

        Parameters
        ----------
        polygon : Polygon
            Input polygon to be used for geometry object
        crs : CRS or str
            The CRS of the input `Polygon`
        """

        assert isinstance(polygon, Polygon)
        self._polygon = polygon
        super().__init__(crs)

    def get_multipolygon(self, **kwargs : Any) -> MultiPolygon:
        """Returns a `MultiPolygon` from the underlying `shapely` geometry

        Parameters
        ----------
        **kwargs : dict, optional
            Currently unused for this class, needed for generic API
            support

        Returns
        -------
        MultiPolygon
            Multipolygon created from the input single polygon
        """

        return MultiPolygon([self._polygon])

    @property
    def polygon(self):
        """Read-only attribute referencing the underlying `Polygon`"""
        return self._polygon

    @property
    def crs(self):
        """Read-only attribute returning the CRS associated with the geometry"""
        return self._crs


class MultiPolygonGeom(ShapelyGeom):

    def __init__(
            self,
            multipolygon: MultiPolygon,
            crs: Union[CRS, str]
            ) -> None:
        """Initialize a `shapely` based geometry object

        Parameters
        ----------
        multipolygon : MultiPolygon
            Input multipolygon to be used for geometry object
        crs : CRS or str
            The CRS of the input `Polygon`
        """

        assert isinstance(multipolygon, MultiPolygon)
        self._multipolygon = multipolygon
        super().__init__(crs)

    def get_multipolygon(self, **kwargs):
        """Returns the underlying `shapely` MultiPolygon

        Parameters
        ----------
        **kwargs : dict, optional
            Currently unused for this class, needed for generic API
            support

        Returns
        -------
        MultiPolygon
            Underlying `Multipolygon`
        """

        return self._multipolygon

    @property
    def multipolygon(self):
        """Read-only attribute referencing the underlying `MultiPolygon`"""
        return self._multipolygon



"""This module defines geometry collector.
`
Geometry collector objects accepts a list of valid basic `Geom`
inputs and creates an object that merges the results of all the
other types of geometries, e.g. mesh-based and raster-based.

Notes
-----
This enables the user to process multiple DEM without having to worry
about the details of merging the output polygons taken from each DEM.
"""

import os
import logging
import warnings
import tempfile
from numbers import Number
from pathlib import Path
from multiprocessing import cpu_count
from typing import Union, Tuple, Optional, Iterable, List, Any

import pandas as pd
import geopandas as gpd
from pyproj import CRS, Transformer
from shapely.geometry import MultiPolygon, Polygon
from shapely import ops
from shapely.validation import explain_validity

from ocsmesh.mesh import Mesh
from ocsmesh.mesh.base import BaseMesh
from ocsmesh.raster import Raster
from ocsmesh.geom.base import BaseGeom
from ocsmesh.geom.raster import RasterGeom
from ocsmesh.geom.mesh import MeshGeom
from ocsmesh.features.contour import FilledContour, Contour
from ocsmesh.features.patch import Patch
from ocsmesh.ops import combine_geometry

CanCreateSingleGeom = Union[Raster, BaseMesh, Polygon, MultiPolygon]
CanCreateMultipleGeom = Iterable[Union[CanCreateSingleGeom, str]]
CanCreateGeom = Union[CanCreateSingleGeom, CanCreateMultipleGeom]

_logger = logging.getLogger(__name__)

class ContourPatchInfoCollector:
    """Helper class for information related to contour patches"""

    def __init__(self) -> None:
        self._contour_patch_info = []

    def add(self,
            contour_defn: FilledContour,
            patch_defn: Optional[Patch]
            ) -> None:
        """Add contour definition information to the contour collection
        """
        self._contour_patch_info.append((contour_defn, patch_defn))

    def __iter__(self):
        for ctr_defn, ptch_defn in self._contour_patch_info:
            yield ctr_defn, ptch_defn


class GeomCollector(BaseGeom):
    """Geometry object based on multiple input types

    Geometry type that merges the information from different types of
    inputs such as raster, mesh and shapely geometries.

    Attributes
    ----------
    crs : CRSDescriptor
        EPSG:4326. All the inputs are transformed before merging
    multipolygon : MultiPolygon
        Lazily calculated `shapely` (multi)polygon of the geometry

    Methods
    -------
    msh_t(**kwargs)
        Returns the `jigsawpy` vertex-edge representation of the geometry
    get_multipolygon(**kwargs)
        Returns `shapely` object representation of the geometry
    add_patch(...)
        Define local (patch) contour extraction definition from all
        the input data (e.g. raster, mesh, etc.)
    """

    def __init__(
            self,
            in_list: CanCreateMultipleGeom,
            base_mesh: Optional[Mesh] = None,
            zmin: Optional[float] = None,
            zmax: Optional[float] = None,
            nprocs: Optional[int] = None,
            chunk_size: Optional[int] = None,
            overlap: Optional[int] = None,
            verbosity: int = 0,
            base_shape: Optional[Union[Polygon, MultiPolygon]] = None,
            base_shape_crs: Union[str, CRS] = 'EPSG:4326'
            ) -> None:
        """Initialize geometry collector object

        Parameters
        ----------
        in_list : list
            List of objects that from which single geometry object
            can be created. This includes path to a raster or mesh
            file as a string, as well as Raster, Mesh or Polygon
            objects. Note that objects are not copied and currently
            any clipping that happens during processing will affect
            the objects pass to the `GeomCollector` constructor.
        base_mesh : Mesh or None, default=None
            Base mesh to be used for extracting boundaries of the
            domain. If not `None` all the input rasters all clipped
            by `base_mesh` polygon before further processing. This
            is useful for cases where we'd like to locally refine
            features of the domain (domain region >> inputs region)
            or when input rasters are much larger that the domain
            and we'd like to extract contours only within domain
            to save on computation.
        zmin : float or None, default=None
            Minimum elevation for extracting domain.
        zmax : float or None, default=None
            Maximum elevation for extracting domain.
        nprocs: int or None, default=None
            Number of processors to use in parallel parts of the
            collector computation
        chunk_size: int or None, default=None
            Chunk size for windowed calculation on rasters
        overlap: int or None default=None
            Window overlap for windowed calculation on rasters
        verbosity: int, default=0,
            Verbosity of the output
        base_shape: Polygon or MultiPolygon or None, default=None
            Similar to `base_mesh`, but instead of calculating the
            polygon from mesh, directly receive it from the calling
            code.
        base_shape_crs: str or CRS, default='EPSG:4326'
            CRS of the input `base_shape`.
        """

        # TODO: Like hfun collector and ops, later move the geom
        # combine functionality here and just call it from ops instead
        # of the other way around

        # For shapely and potentially mesh geom there's no priority
        # definition, they are just unioned with whatever the rest
        # of the input results in

        # store all the info pass
        # pass dem tmpfile or store address of each series
        # (global, patchcontour) to ops and get results and store
        # in the end combine all results (unary_union)

        # NOTE: Input Hfuns and their Rasters can get modified

        # Check nprocs
        nprocs = -1 if nprocs is None else nprocs
        nprocs = cpu_count() if nprocs == -1 else nprocs


        self._elev_info = {'zmin':zmin, 'zmax': zmax}
        self._nprocs = nprocs
        self._chunk_size = chunk_size
        self._overlap = overlap
        self._geom_list = []

        self._base_shape = base_shape
        self._base_shape_crs = CRS.from_user_input(base_shape_crs)

        # NOTE: Base mesh has to have a crs otherwise MeshGeom throws
        # exception
        self._base_mesh = base_mesh
        self._contour_patch_info_coll = ContourPatchInfoCollector()

        self._type_chk(in_list)

        # TODO: CRS considerations -- geom combine doesn't necessarily
        # return EPSG:4326 (unlike hfun collector msh_t)
        self._crs = 'EPSG:4326'

        for in_item in in_list:
            # Add supports(ext) to each hfun type?

            if isinstance(in_item, BaseGeom):
                geom = in_item

            elif isinstance(in_item, Raster):
                if self._base_shape:
                    clip_shape = self._base_shape
                    if not self._base_shape_crs.equals(in_item.crs):
                        transformer = Transformer.from_crs(
                            self._base_shape_crs, in_item.crs, always_xy=True)
                        clip_shape = ops.transform(
                                transformer.transform, clip_shape)
                    try:
                        in_item.clip(clip_shape)
                    except ValueError as err:
                        # This raster does not intersect shape
                        _logger.debug(err)
                        continue

                elif self._base_mesh:
                    try:
                        in_item.clip(self._base_mesh.get_bbox(crs=in_item.crs))
                    except ValueError as err:
                        # This raster does not intersect shape
                        _logger.debug(err)
                        continue

                geom = RasterGeom(in_item, **self._elev_info)

            elif isinstance(in_item, BaseMesh):
                geom = MeshGeom(in_item)

            elif isinstance(in_item, str):
                if in_item.endswith('.tif'):
                    raster = Raster(in_item)
                    if self._base_shape:
                        clip_shape = self._base_shape
                        if not self._base_shape_crs.equals(raster.crs):
                            transformer = Transformer.from_crs(
                                self._base_shape_crs, raster.crs, always_xy=True)
                            clip_shape = ops.transform(
                                    transformer.transform, clip_shape)
                        try:
                            in_item.clip(clip_shape)
                        except ValueError as err:
                            # This raster does not intersect shape
                            _logger.debug(err)
                            continue

                    elif self._base_mesh:
                        try:
                            raster.clip(self._base_mesh.get_bbox(crs=raster.crs))
                        except ValueError as err:
                            # This raster does not intersect shape
                            _logger.debug(err)
                            continue

                    geom = RasterGeom(raster, **self._elev_info)

                elif in_item.endswith(
                        ('.14', '.grd', '.gr3', '.msh', '.2dm')):
                    geom = MeshGeom(Mesh.open(in_item))

                else:
                    raise TypeError("Input file extension not supported!")

            self._geom_list.append(geom)


    def get_multipolygon(self, **kwargs: Any) -> MultiPolygon:
        """Returns the `shapely` representation of the geometry

        Calculates and returns the `MultiPolygon` representation of
        the geometry.

        Parameters
        ----------
        **kwargs : dict, optional
            Currently unused for this class, needed for generic API
            support

        Returns
        -------
        MultiPolygon
            Calculated and merged polygons from all geometry inputs.

        Notes
        -----
        All calculations are done lazily and the results is **not**
        cached. During this process all the stored contour extraction
        specs are applied on all the inputs and then the resulting
        shapes are merged.

        Calculation for each DEM and feature is stored on disk as
        feather files.  In the last steps are all these feather files
        are combined using the out of core calculation by `GeoPandas`.
        """

        # For now we don't need to do any calculations here, the
        # ops will take care of extracting everything. Later the logic
        # in ops needs to move here (like hfun collector)

        # Since raster geoms are stateless, the polygons should be
        # calculated everytime

        epsg4326 = CRS.from_user_input("EPSG:4326")
        mp = None
        with tempfile.TemporaryDirectory() as temp_dir:
            feather_files = []

            temp_path = Path(temp_dir)

            base_multipoly = None
            if self._base_shape:
                base_multipoly = self._base_shape
                if not self._base_shape_crs.equals(epsg4326):
                    transformer = Transformer.from_crs(
                        self._base_shape_crs, epsg4326, always_xy=True)
                    base_multipoly = ops.transform(
                            transformer.transform, base_multipoly)

            elif self._base_mesh:
                # TODO: Make sure all calcs are in EPSG:4326
                base_multipoly = self._base_mesh.hull.multipolygon()

            feather_files.append(self._extract_global_boundary(
                temp_path, base_multipoly))
            feather_files.extend(self._extract_nonraster_boundary(
                temp_path, base_multipoly))
            feather_files.extend(self._extract_features(
                temp_path, base_multipoly))

            gdf = gpd.GeoDataFrame(columns=['geometry'], crs=epsg4326)
            for f in feather_files:
                gdf = pd.concat(
                    [gdf, gpd.read_feather(f).to_crs(gdf.crs)]
                )

            mp = gdf.unary_union
            if isinstance(mp, Polygon):
                mp = MultiPolygon([mp])

            elif not isinstance(mp, MultiPolygon):
                raise ValueError(
                    "Union of all shapes resulted in invalid geometry"
                    + " type")

        return mp

    def add_patch(
            self,
            shape: Optional[Union[MultiPolygon, Polygon]] = None,
            level: Optional[Union[Tuple[float, float], float]] = None,
            contour_defn: Optional[Union[FilledContour, Contour]] = None,
            patch_defn: Optional[Patch] = None,
            shapefile: Optional[Union[str, Path]] = None,
            ) -> None:

        """Specifies an area of geometry inputs to extract contours

        Specifies a localized area of geometry inputs to extract
        contours from.

        Parameters
        ----------
        shape : MultiPolygon or Polygon or None, default=None
            `shapely` (multi)polygon to specify the area from which
            contour specification for this patch should be extracted.
            CRS is assumed to be ``EPSG:4326``.
        level: float or tuple of float, default=None
            Contour level specification for this patch. If provided,
            it could either be a single floating point number for
            the maximum elevation or a tuple indicating minimum and
            maximum elevations for contour extraction.
        contour_defn : FilledContour or Contour or None, default=None
            Alternative way of specifying maximum or minimum or both
            elevation levels for contour extraction.
        patch_defn : Patch or None, default=None
            Alternative way to specify which region within the input
            the contours need to be extracted from.
        shapefile: str or Path or None, default=None
            Alternative way to specify which region within the input
            the contours need to be extracted from.

        Returns
        -------
        None

        Notes
        -----
        This method doesn't result in any calculations. It simply
        stores the information about the user specified contour.
        During the `get_multipolygon` call the contours are actually
        calculated on all the inputs.
        """

        # Always lazy

        if not contour_defn:
            level0 = None
            level1 = None
            if isinstance(level, tuple) and len(level) == 2:
                level0 = level[0]
                level1 = level[1]
            elif isinstance(level, Number):
                level0 = None
                level1 = level
            else:
                raise ValueError(
                    "Level must be specified either by min and max values"
                    " or by only max value ")

            contour_defn = FilledContour(level1=level)

        elif isinstance(contour_defn, Contour):
            contour_defn = FilledContour(max_contour_defn=contour_defn)

        elif not isinstance(contour_defn, FilledContour):
            raise TypeError(
                f"Filled contour definition must be of type"
                f" {FilledContour} not {type(contour_defn)}!")

        elif level is not None:
            msg = "Level is ignored since a contour definition is provided!"
            warnings.warn(msg)
            _logger.info(msg)

        if not patch_defn:
            if shape:
                patch_defn = Patch(shape=shape)

            elif shapefile:
                patch_defn = Patch(shapefile=shapefile)

        elif not isinstance(patch_defn, Patch):
            raise TypeError(
                f"Patch definition must be of type {Patch} not"
                f" {type(patch_defn)}!")


        # If patch defn is None it means the patch applies to
        # all the sources of the accompanying contour
        self._contour_patch_info_coll.add(
            contour_defn, patch_defn)


    def _type_chk(self, input_list) -> None:
        """Checks the if the input types are supported for geometry

        Checks if geometry collector supports handling geometry created
        from the input types.

        Parameters
        ----------
        input_list : List[Any]

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If the any of the inputs are not supported
        """

        valid_types = (str, Raster, BaseGeom, BaseMesh)
        if not all(isinstance(item, valid_types) for item in input_list):
            raise TypeError(
                f'Input list items must be of type {", ".join(valid_types)}'
                f', or a derived type.')

    def _get_raster_sources(self) -> List[Union[RasterGeom, Raster]]:
        """Get the list rasters from inputs to the object constructor

        Retruns
        -------
        list of RasterGeom or Raster
            Filtered input list to return only rasters
        """

        raster_types = (RasterGeom, Raster)
        rasters = [
            i for i in self._geom_list if isinstance(i, raster_types)]
        return rasters

    def _get_raster_source_files(self) -> List[Path]:
        """Get the list of raster temporary files

        Returns
        -------
        list of path-like
            List of path to geom input raster object temporary files.
        """

        rasters = self._get_raster_sources()
        return self._get_raster_files_from_source(rasters)

    def _get_raster_files_from_source(
            self,
            rasters: Iterable[Union[Raster, RasterGeom]]
            ) -> List[Path]:
        """Get the list of raster temporary files

        Get the list of raster temporary files for input argument
        `rasters`.

        Parameters
        ----------
        rasters : iterable
            List of raster objects or geometries for which the
            temporary file address is returned

        Returns
        -------
        list of path-like
            List of path to method input raster object temporary files.
        """

        raster_files = []
        for r in rasters:
            if isinstance(r, Raster):
                raster_files.append(r.tmpfile)
            elif isinstance(r, RasterGeom):
                raster_files.append(r.raster.tmpfile)

        return raster_files

    def _get_non_raster_sources(self) -> List[Any]:
        """Get the list non-rasters from inputs in the constructor

        Retruns
        -------
        list
            Filtered input list to return only non-rasters
        """
        raster_types = (RasterGeom, Raster)
        non_rasters = [
            i for i in self._geom_list if not isinstance(i, raster_types)]
        return non_rasters

    def _get_valid_multipolygon(
            self,
            polygon: Union[Polygon, MultiPolygon]
            ) -> MultiPolygon:
        """Get a valid multipolygon from the input `polygon`

        Validates and if applicable creates a multipolygon from the
        input argument `polygon`

        Parameters
        ----------
        polygon : Polygon or MultiPolygon
            The input polygon or multipolygon which might not be
            topologically valid.

        Returns
        -------
        MultiPolygon
            A validated `shapely` `MultiPolygon` entity
        """

        # TODO: Performance bottleneck for valid checks
        if not polygon.is_valid:
            polygon = ops.unary_union(polygon)

            if not polygon.is_valid:
                polygon = polygon.buffer(0)

            if not polygon.is_valid:
                raise ValueError(explain_validity(polygon))

        if isinstance(polygon, Polygon):
            polygon = MultiPolygon([polygon])

        return polygon

    def _extract_global_boundary(
            self,
            out_dir: Union[Path, str],
            base_multipoly: Optional[MultiPolygon]
            ) -> Path:
        """Calculates the final boundary from all the raster inputs

        Parameters
        ----------
        out_dir : str
            Output directory into which feather files should be written
        base_multipoly : MultiPolygon
            Base shape to use for clipping DEM data.

        Returns
        -------
        path-like
            Path to a feather file containing the final boundary
            based on raster inputs
        """

        out_path = Path(out_dir)

        geom_path = out_path / 'global_boundary.feather'

        raster_files = self._get_raster_source_files()
        zmin = self._elev_info['zmin']
        zmax = self._elev_info['zmax']
        _logger.info("Extracting global boundaries")
        combine_geometry(
            raster_files, geom_path, "feather",
            None, base_multipoly, False,
            zmin, zmax,
            self._chunk_size, self._overlap,
            self._nprocs)

        return geom_path

    def _extract_nonraster_boundary(
            self,
            out_dir: Union[Path, str],
            base_multipoly: Optional[MultiPolygon]
            ) -> Path:
        """Calculates the final boundary from all the non-raster inputs

        Parameters
        ----------
        out_dir : str
            Output directory into which feather files should be written
        base_multipoly : MultiPolygon
            Base shape to use for clipping elevation data mesh.

        Returns
        -------
        path-like
            Path to a feather file containing the final boundary
            based on non-raster inputs
        """

        out_path = Path(out_dir)

        non_rasters = self._get_non_raster_sources()
        feather_files = []
        for e, geom in enumerate(non_rasters):

            geom_path = out_path / f'nonraster_{os.getpid()}_{e}.feather'

            crs = geom.crs
            multipoly = self._get_valid_multipolygon(
                    geom.get_multipolygon())
            gdf_non_raster = gpd.GeoDataFrame(
                    {'geometry': multipoly}, crs=crs)
            if crs != CRS.from_user_input("EPSG:4326"):
                gdf_non_raster = gdf_non_raster.to_crs("EPSG:4326")

            # TODO: Clip using base_multipoly?

            gdf_non_raster.to_feather(geom_path)

            feather_files.append(geom_path)

        return feather_files

    def _extract_features(
            self,
            out_dir: Union[Path, str],
            base_multipoly: Optional[MultiPolygon]
            ) -> List[Path]:
        """Calculates the local feature polygons

        Parameters
        ----------
        out_dir : str
            Output directory into which feather files should be written
        base_multipoly : MultiPolygon
            Base shape to use for clipping elevation data.

        Returns
        -------
        list of path-like
            Paths to feather files containing all the extracted features

        Notes
        -----
        Currently the only feature available is patch feature extraction
        """


        feather_files = []
        feather_files.extend(self._apply_patch(out_dir, base_multipoly))

        return feather_files

    def _apply_patch(
            self,
            out_dir: Union[Path, str],
            base_multipoly: Optional[MultiPolygon]
            ) -> List[Path]:
        """Extracts contours based on patch specifications

        Extracts the domain filled contours based on the specifications
        user provided earlier by `add_patch`

        Parameters
        ----------
        out_dir : str
            Output directory into which feather files should be written
        base_multipoly : MultiPolygon
            Base shape to use for clipping elevation data.

        Returns
        -------
        list of path-like
            List of extracted contours stored on files on the disk
        """

        out_path = Path(out_dir)


        raster_files = self._get_raster_source_files()
        zmin = self._elev_info['zmin']
        zmax = self._elev_info['zmax']

        feather_files = []
        for e, (ctr_defn, ptch_defn) in enumerate(self._contour_patch_info_coll):

            patch_zmin, patch_zmax = ctr_defn.level
            if not patch_zmin:
                patch_zmin = zmin
            if not patch_zmax:
                patch_zmax = zmax

            patch_raster_files = raster_files
            if ctr_defn.has_source:
                patch_rasters = ctr_defn.sources
                patch_raster_files = self._get_raster_files_from_source(
                        patch_rasters)

            # Pass patch shape instead of base mesh
            # See explanation in add_patch
            _logger.info("Extracting patch contours")
            combine_poly = base_multipoly
            if ptch_defn:
                patch_mp, crs = ptch_defn.get_multipolygon()
                gdf_patch = gpd.GeoDataFrame(
                        {'geometry': patch_mp}, crs=crs)
                if crs != CRS.from_user_input("EPSG:4326"):
                    gdf_patch = gdf_patch.to_crs("EPSG:4326")
                combine_poly = MultiPolygon(list(gdf_patch.geometry))
            geom_path = out_path / f'patch_{os.getpid()}_{e}.feather'
            combine_geometry(
                patch_raster_files, geom_path, "feather",
                None, combine_poly, True,
                patch_zmin, patch_zmax,
                self._chunk_size, self._overlap,
                self._nprocs)

            if geom_path.is_file():
                feather_files.append(geom_path)

        return feather_files



import pathlib
from importlib import util
import tempfile
import os
import sys
import platform


try:
    import jigsawpy  # noqa: F401
except OSError as e:
    pkg = util.find_spec("jigsawpy")
    libjigsaw = {
            "Windows": "jigsaw.dll",
            "Linux": "libjigsaw.so",
            "Darwin": "libjigsaw.dylib"
            }[platform.system()]
    tgt_libpath = pathlib.Path(pkg.origin).parent / "_lib" / libjigsaw
    pyenv = pathlib.Path("/".join(sys.executable.split('/')[:-2]))
    src_libpath = pyenv / 'lib' / libjigsaw
    if not src_libpath.is_file():
        raise e

    os.symlink(src_libpath, tgt_libpath)


from .raster import Raster
from .mesh import Mesh
from .geom import Geom
from .hfun import Hfun
from .driver import JigsawDriver

if util.find_spec("colored_traceback") is not None:
    import colored_traceback
    colored_traceback.add_hook(always=True)

tmpdir = str(pathlib.Path(tempfile.gettempdir()+'/ocsmesh'))+'/'
os.makedirs(tmpdir, exist_ok=True)

__all__ = [
    "Geom",
    "Hfun",
    "Raster",
    "Mesh",
    "JigsawDriver",
]

# mpl.rcParams['agg.path.chunksize'] = 10000



import logging
import os
import pathlib
from multiprocessing import cpu_count
from typing import Union, Sequence, List

from pyproj import CRS
from jigsawpy import savemsh, savevtk

from ocsmesh.raster import Raster
from ocsmesh.hfun.hfun import Hfun
from ocsmesh.mesh.mesh import Mesh
from ocsmesh import utils

_logger = logging.getLogger(__name__)


class HfunCombine:

    def __init__(
            self,
            dem_files: Sequence[Union[str, os.PathLike]],
            out_file: Union[str, os.PathLike],
            out_format: str = "shapefile",
            mesh_file: Union[str, os.PathLike, None] = None,
            hmin: Union[float, None] = None,
            hmax: Union[float, None] = None,
            contours: List[List[float]] = None,
            constants: List[List[float]] = None,
            chunk_size: Union[int, None] = None,
            overlap: Union[int, None] = None,
            method: str = 'exact',
            nprocs: int = -1):


        self._base_exterior = None

        self._operation_info = {
                'dem_files': dem_files,
                'out_file': out_file,
                'out_format': out_format,
                'mesh_file': mesh_file,
                'hmin': hmin,
                'hmax': hmax,
                'contours': contours,
                'constants': constants,
                'chunk_size': chunk_size,
                'overlap': overlap,
                'method': method,
                'nprocs': nprocs
        }

    def run(self):

        dem_files = self._operation_info['dem_files']
        out_file = self._operation_info['out_file']
        out_format = self._operation_info['out_format']
        mesh_file = self._operation_info['mesh_file']
        hmin = self._operation_info['hmin']
        hmax = self._operation_info['hmax']
        contours = self._operation_info['contours']
        constants = self._operation_info['constants']
        chunk_size = self._operation_info['chunk_size']
        overlap = self._operation_info['overlap']
        method = self._operation_info['method']
        nprocs = self._operation_info['nprocs']

        nprocs = cpu_count() if nprocs == -1 else nprocs

        out_dir = pathlib.Path(out_file).parent
        out_dir.mkdir(exist_ok=True, parents=True)

        logging.info("Loading base mesh...")
        base_mesh = None
        if mesh_file:
            base_mesh = Mesh.open(mesh_file, crs="EPSG:4326")

        rast_list = []
        for dem_path in dem_files:

            logging.info(f"Loading raster {dem_path}...")
            rast = Raster(
                    dem_path, chunk_size=chunk_size, overlap=overlap)
            rast_list.append(rast)

        # Create Hfun
        logging.info("Creating Hfun from rasters...")
        hfun_collector = Hfun(
                rast_list, base_mesh=base_mesh,
                hmin=hmin, hmax=hmax, nprocs=nprocs, method=method)

        for contour in contours:
            logging.info("Adding contour refinement...")
            if len(contour) > 3:
                raise ValueError(
                    "Invalid format for contour specification."
                    " It should be level [expansion target-size].")

            level, expansion_rate, target_size = [
                    *contour, *[None]*(3-len(contour))]

            if level is None:
                raise ValueError(
                    "Invalid format for contour specification."
                    " It should be level [expansion target-size].")
            if expansion_rate is None:
                expansion_rate = 0.1
            if target_size is None:
                target_size = hmin

            hfun_collector.add_contour(
                level, expansion_rate, target_size)

        for lower_bound, target_size in constants:
            hfun_collector.add_constant_value(
                    value=target_size, lower_bound=lower_bound)

        self._write_to_file(
                out_format, out_file, hfun_collector, 'EPSG:4326')


    def _write_to_file(
            self, out_format, out_file, hfun_collector, crs):

        _logger.info(f"Writing for file ({out_format}) ...")

        # NOTE: Combined mesh from collector is always in EPSG:4326
        jig_hfun = hfun_collector.msh_t()
        dst_crs = CRS.from_user_input(crs)
        if jig_hfun.crs != dst_crs:
            _logger.info(f"Reprojecting hfun to ({crs}) ...")
            utils.reproject(jig_hfun, dst_crs)

        # TODO: Check for correct extension on out_file
        if out_format in ("jigsaw", "vtk"):
            if out_format == "jigsaw":
                savemsh(out_file, jig_hfun)

            elif out_format == "vtk":
                savevtk(out_file, jig_hfun)

        elif out_format in ['2dm', 'sms']:
            # TODO: How to specify crs in 2dm file?
            mesh = Mesh(jig_hfun)
            mesh.write(out_file, format='2dm')

        else:
            raise NotImplementedError(f"Output type {out_format} is not supported")

        _logger.info("Done")



#! python
import unittest
from copy import deepcopy
from pathlib import Path
import shutil
import tempfile

from jigsawpy import jigsaw_msh_t
import geopandas as gpd
import numpy as np
from pyproj import CRS
import rasterio as rio
import requests
from shapely import geometry

import ocsmesh


# TODO: Move these helper functions to `utils`
def _helper_raster_from_numpy(
    filename,
    data,
    mgrid,
    crs=CRS.from_epsg(4326)
):
    x = mgrid[0][:, 0]
    y = mgrid[1][0, :]
    res_x = (x[-1] - x[0]) / data.shape[1]
    res_y = (y[-1] - y[0]) / data.shape[0]
    transform = rio.transform.Affine.translation(
        x[0], y[0]
    ) * rio.transform.Affine.scale(res_x, res_y)
    if not isinstance(crs, CRS):
        crs = CRS.from_user_input(crs)

    with rio.open(
        filename,
        'w',
        driver='GTiff',
        height=data.shape[0],
        width=data.shape[1],
        count=1,
        dtype=data.dtype,
        crs=crs,
        transform=transform,
    ) as dst:
        dst.write(data, 1)


def _helper_msh_t_from_numpy(
    coordinates,
    triangles,
    crs=CRS.from_epsg(4326)
):
    if not isinstance(crs, CRS):
        crs = CRS.from_user_input(crs)
    mesh = jigsaw_msh_t()
    mesh.mshID = 'euclidean-mesh'
    mesh.ndims = +2
    mesh.crs = crs
    mesh.vert2 = np.array(
        [(coord, 0) for coord in coordinates],
        dtype=jigsaw_msh_t.VERT2_t
        )
    mesh.tria3 = np.array(
        [(index, 0) for index in triangles],
        dtype=jigsaw_msh_t.TRIA3_t
        )

    return mesh



class SizeFromMesh(unittest.TestCase):

    def setUp(self):
        rast = ocsmesh.raster.Raster('/tmp/test_dem.tif')

        hfun_orig = ocsmesh.hfun.hfun.Hfun(rast, hmin=100, hmax=1500)
        hfun_orig.add_contour(level=0, expansion_rate=0.001, target_size=100)
        hfun_orig_jig = hfun_orig.msh_t()

        self.hfun_orig_val = hfun_orig_jig.value

        hfun_calc_jig = deepcopy(hfun_orig_jig)
        mesh_calc = ocsmesh.mesh.mesh.Mesh(hfun_calc_jig)
        self.hfun_calc = ocsmesh.hfun.hfun.Hfun(mesh_calc)
        self.hfun_calc.size_from_mesh()

    def test_calculated_size(self):
        hfun_calc_jig = self.hfun_calc.msh_t()

        hfun_calc_val = hfun_calc_jig.value
        hfun_val_diff = self.hfun_orig_val - hfun_calc_val

        # TODO: Come up with a more robust criteria
        threshold = 0.2
        err_value = np.max(np.abs(hfun_val_diff))/np.max(self.hfun_orig_val)
        self.assertTrue(err_value < threshold)



class CourantNumConstraint(unittest.TestCase):

    def test_calculate_element_size_from_courant_num(self):
        self.assertEqual(
            ocsmesh.utils.get_element_size_courant(
                target_courant=1,
                characteristic_velocity_magnitude=1,
                timestep=1
            ),
            1
        )

        self.assertEqual(
            ocsmesh.utils.get_element_size_courant(
                target_courant=0.5,
                characteristic_velocity_magnitude=2,
                timestep=3.5
            ),
            14
        )

        self.assertTrue(
            np.array_equal(
                ocsmesh.utils.get_element_size_courant(
                    target_courant=1,
                    characteristic_velocity_magnitude=np.array([1, 1]),
                    timestep=1
                ),
                np.array([1, 1])
            )
        )

        # TODO: Add non-trivial cases


    def test_approx_courant_number_for_depth(self):
        elemsize = np.ones((10, 20)) * 1
        depths = -np.arange(1, 201).reshape(elemsize.shape)
        timestep = 1
        wave_amplitude = 2

        C_apprx = ocsmesh.utils.approximate_courant_number_for_depth(
            depths, timestep, elemsize, wave_amplitude
        )

        self.assertTrue(np.all(C_apprx > 0))

        # TODO: Add non-trivial cases


    def test_can_velocity_be_approx_line_wave_theory(self):
        self.assertFalse(
            ocsmesh.utils.can_velocity_be_approximated_by_linear_wave_theory(1, 2)
        )

        self.assertFalse(
            ocsmesh.utils.can_velocity_be_approximated_by_linear_wave_theory(-1, 2)
        )

        self.assertTrue(
            ocsmesh.utils.can_velocity_be_approximated_by_linear_wave_theory(-3, 2)
        )

        self.assertTrue(
            ocsmesh.utils.can_velocity_be_approximated_by_linear_wave_theory(-2)
        )

        self.assertFalse(
            ocsmesh.utils.can_velocity_be_approximated_by_linear_wave_theory(-1.9)
        )

        self.assertTrue(
            np.array_equal(
                ocsmesh.utils.can_velocity_be_approximated_by_linear_wave_theory(
                    np.array([-1, -1, 0, 1, -2], dtype=float),
                    1
                ),
                np.array([True, True, False, False, True])
            )
        )


    def test_estimate_velocity_magnitude_for_depth(self):
        # Velocity approx for depths shallower than wave amp is the same
        self.assertEqual(
            ocsmesh.utils.estimate_particle_velocity_from_depth(-1),
            ocsmesh.utils.estimate_particle_velocity_from_depth(0)
        )
        self.assertEqual(
            ocsmesh.utils.estimate_particle_velocity_from_depth(-1, 3),
            ocsmesh.utils.estimate_particle_velocity_from_depth(0, 3)
        )

        self.assertNotEqual(
            ocsmesh.utils.estimate_particle_velocity_from_depth(-2, 3),
            ocsmesh.utils.estimate_particle_velocity_from_depth(-4, 3)
        )

        # TODO: Should tests for exact approx value matches be added here?


    def test_cfl_constraint_object_api_type(self):
        sizes_before_constraint = np.array([[500, 600]])
        depths = np.array([[-10, -11]])
        constraint = ocsmesh.features.constraint.CourantNumConstraint(
            value=0.5
        )
        sizes_after_constraint = constraint.apply(depths, sizes_before_constraint)

        # Assertions
        self.assertIsInstance(
            sizes_after_constraint, type(sizes_before_constraint)
        )

    def test_cfl_constraint_object_api_io(self):
        sizes_before_constraint = np.array([[500, 600]])
        depths_1 = np.array([[-10, -11]])
        depths_2 = np.array([[-10]])
        constraint = ocsmesh.features.constraint.CourantNumConstraint(
            value=0.5
        )

        # Assertions
        self.assertEqual(
            constraint.apply(depths_1, sizes_before_constraint).shape,
            sizes_before_constraint.shape
        )
        self.assertRaises(
            ValueError,
            constraint.apply, depths_2, sizes_before_constraint,
        )


    def test_cfl_constraint_max(self):
        sizes_before_constraint = np.ones((10, 20)) * np.finfo(float).resolution
        depths = np.ones((10, 20)) * -20

        constraint = ocsmesh.features.constraint.CourantNumConstraint(
            value=0.5,
            value_type='max',
            wave_amplitude=2,
            timestep=150,
        )

        sizes_after_constraint = constraint.apply(depths, sizes_before_constraint)

        # Assertions
        # NOTE: Max is on Courant # NOT the element size
        self.assertTrue(
            np.all(sizes_after_constraint > sizes_before_constraint)
        )


    def test_cfl_constraint_min(self):
        sizes_before_constraint = np.ones((10, 20)) * np.inf
        depths = np.ones((10, 20)) * -20

        constraint = ocsmesh.features.constraint.CourantNumConstraint(
            value=0.1,
            value_type='min',
            wave_amplitude=2,
            timestep=150,
        )

        sizes_after_constraint = constraint.apply(depths, sizes_before_constraint)

        # Assertions
        # NOTE: Max is on Courant # NOT the element size
        self.assertTrue(
            np.all(sizes_after_constraint < sizes_before_constraint)
        )


    def test_cfl_constraint_unaffected(self):
        sizes_before_constraint = np.vstack(
            (
                np.ones((10, 20)) * np.finfo(float).resolution,
                np.ones((10, 20)) * np.inf
            )
        )
        depths = -np.arange(1, 401).reshape(sizes_before_constraint.shape)

        constraint = ocsmesh.features.constraint.CourantNumConstraint(
            value=0.1,
            value_type='min',
            wave_amplitude=2,
            timestep=150,
        )

        sizes_after_constraint = constraint.apply(depths, sizes_before_constraint)
        self.assertTrue(
            np.all(
                np.equal(
                    sizes_after_constraint.ravel()[:200],
                    sizes_before_constraint.ravel()[:200]
                )
            )
        )


    def test_cfl_constraint_result(self):
        sizes_before_constraint = np.ones((10, 20)) * np.inf
        depths = -np.arange(1, 201).reshape(sizes_before_constraint.shape)

        constraint = ocsmesh.features.constraint.CourantNumConstraint(
            value=0.1,
            value_type='min',
            wave_amplitude=2,
            timestep=150,
        )

        sizes_after_constraint = constraint.apply(depths, sizes_before_constraint)

        C_apprx = ocsmesh.utils.approximate_courant_number_for_depth(
            depths, 150, sizes_after_constraint, 2
        )

        # Assertions
        self.assertTrue(np.all(np.isclose(C_apprx, 0.1)))


class SizeFunctionWithCourantNumConstraint(unittest.TestCase):

    # NOTE: Since mesh size function doesn't have built-in depth
    # information, just like other constraints Courant number constraint
    # is not implemented for it either!

    def test_hfun_raster_cfl_constraint_support(self):
        dt = 100
        nu = 2
        courant_hi = 0.8
        courant_lo = 0.2

        rast = ocsmesh.raster.Raster('/tmp/test_dem.tif')

        hfun_raster = ocsmesh.hfun.hfun.Hfun(rast, hmin=100, hmax=5000)
        hfun_raster.add_courant_num_constraint(
            upper_bound=courant_hi,
            lower_bound=courant_lo,
            timestep=dt,
            wave_amplitude=nu
        )

        C_apprx = ocsmesh.utils.approximate_courant_number_for_depth(
            rast.values, dt, hfun_raster.values, nu
        )

        self.assertTrue(
            np.all(
                np.logical_and(
                    np.logical_or(
                        C_apprx > courant_lo,
                        np.isclose(C_apprx, courant_lo)
                    ),
                    np.logical_or(
                        C_apprx < courant_hi,
                        np.isclose(C_apprx, courant_hi)
                    )
                )
            )
        )

        hfun_jig = hfun_raster.msh_t()
        mesh_jig = deepcopy(hfun_jig)
        mesh = ocsmesh.mesh.mesh.Mesh(mesh_jig)
        mesh.interpolate(rast, nprocs=1)

        C_apprx_mesh = ocsmesh.utils.approximate_courant_number_for_depth(
            mesh.msh_t.value, dt, hfun_jig.value, nu
        )

        # Note using higher tolerance for closeness since sizes and
        # depths are interpolated. Is this ideal? No!
        self.assertTrue(
            np.all(
                np.logical_and(
                    np.logical_or(
                        C_apprx_mesh > courant_lo,
                        np.isclose(C_apprx_mesh, courant_lo, atol=0.03)
                    ),
                    np.logical_or(
                        C_apprx_mesh < courant_hi,
                        np.isclose(C_apprx_mesh, courant_hi, atol=0.03)
                    )
                )
            )
        )

    def test_hfun_raster_cfl_constraint_io(self):
        rast = ocsmesh.raster.Raster('/tmp/test_dem.tif')
        hfun_raster = ocsmesh.hfun.hfun.Hfun(rast, hmin=100, hmax=5000)
        self.assertRaises(
            ValueError,
            hfun_raster.add_courant_num_constraint,
            upper_bound=None,
            lower_bound=None,
            timestep=100,
            wave_amplitude=2
        )


    def test_hfun_coll_cfl_constraint(self):
        dt = 100
        nu = 2
        courant_hi = 0.90
        courant_lo = 0.2

        # Fast method is much less accurate!
        method_tolerance = {
            'exact': 0.03,
            'fast': 0.2
        }

        for method, tol in method_tolerance.items():
            # TODO: Add subTest

            # Creating adjacent rasters from the test raster
            rast1 = ocsmesh.raster.Raster('/tmp/test_dem.tif')
            rast2 = ocsmesh.raster.Raster('/tmp/test_dem.tif')
            bounds = rast1.bbox.bounds
            bbox1 = geometry.box(
                bounds[0], bounds[1], (bounds[0] + bounds[2]) / 2, bounds[3]
            )
            bbox2 = geometry.box(
                (bounds[0] + bounds[2]) / 2, bounds[1], bounds[2], bounds[3]
            )
            rast1.clip(bbox1)
            rast2.clip(bbox2)

            hfun_coll = ocsmesh.hfun.hfun.Hfun(
                [rast1, rast2], hmin=100, hmax=5000, method=method
            )
            hfun_coll.add_courant_num_constraint(
                upper_bound=courant_hi,
                lower_bound=courant_lo,
                timestep=dt,
                wave_amplitude=nu
            )

            hfun_jig = hfun_coll.msh_t()
            mesh_jig = deepcopy(hfun_jig)
            mesh = ocsmesh.mesh.mesh.Mesh(mesh_jig)
            mesh.interpolate([rast1, rast2], method='nearest', nprocs=1)

            C_apprx_mesh = ocsmesh.utils.approximate_courant_number_for_depth(
                mesh.msh_t.value, dt, hfun_jig.value, nu
            )

            valid_courant = np.logical_and(
                np.logical_or(
                    C_apprx_mesh > courant_lo,
                    np.isclose(C_apprx_mesh, courant_lo, atol=tol)
                ),
                np.logical_or(
                    C_apprx_mesh < courant_hi,
                    np.isclose(C_apprx_mesh, courant_hi, atol=tol)
                )
            )
            # Note using higher tolerance for closeness since sizes and
            # depths are interpolated. Is this ideal? No!
            self.assertTrue(
                np.all(valid_courant),
                msg=f"Courant constraint failed for '{method}' method!"
                    + f"\n{C_apprx_mesh[~valid_courant]}"
            )


class SizeFunctionCollectorAddFeature(unittest.TestCase):

    def setUp(self):
        self.tdir = Path(tempfile.mkdtemp())

        self.rast1 = self.tdir / 'rast_1.tif'
        self.rast2 = self.tdir / 'rast_2.tif'
        self.mesh1 = self.tdir / 'mesh_1.gr3'
        self.feat1 = self.tdir / 'feature_1'
        self.feat2 = self.tdir / 'feature_2'

        rast_xy_1 = np.mgrid[-1:0.1:0.1, -0.7:0.1:0.1]
        rast_xy_2 = np.mgrid[0:1.1:0.1, -0.7:0.1:0.1]
        rast_z_1 = np.ones_like(rast_xy_1[0])

        _helper_raster_from_numpy(
            self.rast1, rast_z_1, rast_xy_1, 4326
        )
        _helper_raster_from_numpy(
            self.rast2, rast_z_1, rast_xy_2, 4326
        )

        crd = np.array([
            [-1, 0],
            [-0.1, 0],
            [-0.5, 0.2],
            [0.1, 0],
            [1, 0],
            [0.5, 0.2],
            [-1, 0.4],
            [1, 0.4],
        ])
        tria = np.array([
            [0, 1, 2],
            [3, 4, 5],
            [2, 1, 3],
            [2, 3, 5],
            [2, 6, 7],
            [2, 7, 5],
            [0, 2, 6],
            [5, 4, 7],
        ])
        msh_t = _helper_msh_t_from_numpy(crd, tria, 4326)
        mesh = ocsmesh.Mesh(msh_t)
        mesh.write(str(self.mesh1), format='grd', overwrite=False)

        self.shape1 = geometry.LineString([
            [-1, 0], [1, 0]
        ])

        self.shape2 = geometry.LineString([
            [0, -1], [0, 1]
        ])

        gdf_feature = gpd.GeoDataFrame(geometry=[self.shape1], crs=4326)
        gdf_feature.to_file(self.feat1)

        gdf_feature = gpd.GeoDataFrame(geometry=[self.shape2], crs=4326)
        gdf_feature.to_file(self.feat2)

    def tearDown(self):
        shutil.rmtree(self.tdir)

    def _create_hfun(self, hmin, hmax):

        rast1 = ocsmesh.Raster(self.rast1)
        rast2 = ocsmesh.Raster(self.rast2)

        hfun_mesh_1 = ocsmesh.Hfun(ocsmesh.Mesh.open(self.mesh1, crs=4326))
        hfun_mesh_1.size_from_mesh()

        hfun = ocsmesh.Hfun(
            [rast1, rast2, hfun_mesh_1],
            hmin=hmin,
            hmax=hmax,
            method='exact')

        return hfun

    def _check_applied_refinement(self, msh_t, refine_gdf, target_size):

        refine_msh_t = ocsmesh.Hfun(ocsmesh.Mesh(ocsmesh.utils.clip_mesh_by_shape(
            mesh=msh_t,
            shape=refine_gdf.unary_union,
            fit_inside=True,
            inverse=False
        )))
        refine_msh_t.size_from_mesh()
        refine_avg = np.mean(refine_msh_t.msh_t().value)
        rest_msh_t = ocsmesh.Hfun(ocsmesh.Mesh(ocsmesh.utils.clip_mesh_by_shape(
            mesh=msh_t,
            shape=refine_gdf.unary_union,
            fit_inside=False,
            inverse=True
        )))
        rest_msh_t.size_from_mesh()
        rest_avg = np.mean(rest_msh_t.msh_t().value)

        self.assertTrue(np.isclose(refine_avg, target_size, rtol=3e-1))
        self.assertTrue(rest_avg > target_size * 10)

    def _is_refined_by_shape1(self, hfun, target_size):

        hfun_msh_t = hfun.msh_t()

        # Nodes close to the feature line must be small
        gdf_feature = gpd.GeoDataFrame(
            geometry=[self.shape1], crs=4326
        )
        gdf_clip = gdf_feature.to_crs(hfun_msh_t.crs).buffer(target_size*1.1)
        self._check_applied_refinement(hfun_msh_t, gdf_clip, target_size)


    def _is_refined_by_feat1(self, hfun, target_size):

        hfun_msh_t = hfun.msh_t()

        # Nodes close to the feature line must be small
        gdf_feature = gpd.read_file(self.feat1)
        gdf_clip = gdf_feature.to_crs(hfun_msh_t.crs).buffer(target_size*1.1)
        self._check_applied_refinement(hfun_msh_t, gdf_clip, target_size)

    def _is_refined_by_feat2(self, hfun, target_size):

        hfun_msh_t = hfun.msh_t()

        # Nodes close to the feature line must be small
        gdf_feature = gpd.read_file(self.feat2)
        gdf_clip = gdf_feature.to_crs(hfun_msh_t.crs).buffer(target_size)
        self._check_applied_refinement(hfun_msh_t, gdf_clip, target_size)



    def test_by_shape(self):

        hmin = 500
        hmax = 10000

        hfun = self._create_hfun(hmin, hmax)

        hfun.add_feature(
            shape=self.shape1,
            expansion_rate=0.002,
            target_size=hmin,
            crs=4326
        )
        self._is_refined_by_shape1(hfun, hmin)


    def test_by_shapefile(self):

        hmin = 500
        hmax = 10000

        hfun = self._create_hfun(hmin, hmax)

        hfun.add_feature(
            shapefile=self.feat1,
            expansion_rate=0.002,
            target_size=hmin,
        )

        self._is_refined_by_feat1(hfun, hmin)


    def test_by_linedefn(self):

        hmin = 500
        hmax = 10000

        hfun = self._create_hfun(hmin, hmax)

        hfun.add_feature(
            line_defn=ocsmesh.features.linefeature.LineFeature(shapefile=self.feat1),
            expansion_rate=0.002,
            target_size=hmin,
        )

        self._is_refined_by_feat1(hfun, hmin)


    def test_optional_arg_priority(self):

        hmin = 500
        hmax = 10000

        # First priority is `line_defn`
        hfun1 = self._create_hfun(hmin, hmax)
        hfun1.add_feature(
            line_defn=ocsmesh.features.linefeature.LineFeature(shapefile=self.feat1),
            shape=self.shape2,
            crs=4326,
            shapefile=self.feat2,
            expansion_rate=0.002,
            target_size=hmin,
        )
        self._is_refined_by_feat1(hfun1, hmin)


        # Second priority is `shape` + `crs`
        hfun2 = self._create_hfun(hmin, hmax)
        hfun2.add_feature(
            shape=self.shape1,
            crs=4326,
            shapefile=self.feat2,
            expansion_rate=0.002,
            target_size=hmin,
        )
        self._is_refined_by_feat1(hfun2, hmin)


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from abc import ABC, abstractmethod

import numpy as np
from scipy.spatial import cKDTree
from scipy import constants

from ocsmesh import utils

ConstraintValueType = Enum("ConstraintValueType", "MIN MAX")

class Constraint(ABC):

    def __init__(self, value_type: str = 'min', rate: float = 0.1):

        # TODO: Put rate in a mixin ?
        self._rate = rate

        # pylint: disable=W1116
        if isinstance(value_type, ConstraintValueType):
            self._val_type = value_type
        elif isinstance(value_type, str):
            self._val_type = ConstraintValueType[value_type.upper()]
        else:
            raise ValueError("Invalid input for value type!")


    @property
    def type(self):
        return type(self)


    @property
    def value_type(self):
        return self._val_type

    @property
    def satisfies(self):

        '''
        The function to compare a value with the constraint value
        and evaluate whether it satisfies the constraint
        function's needs to receive values to check as first argument
        '''

        # pylint: disable=R1705
        if self.value_type == ConstraintValueType.MIN:
            return np.greater
        elif self.value_type == ConstraintValueType.MAX:
            return np.less

        raise ValueError("Invalid value type for constraint!")

    @property
    def rate_sign(self):
        # TODO: Put this method in a mixin

        '''
        Based on the value-type of the constraints, return a sign
        indicating whether rate is for expansion or contraction of
        size outside calculated zone
        '''

        # pylint: disable=R1705
        if self.value_type == ConstraintValueType.MIN:
            return -1
        elif self.value_type == ConstraintValueType.MAX:
            return +1

        raise ValueError("Invalid value type for constraint!")


    @abstractmethod
    def apply(self):
        pass


    def _apply_rate(self, ref_values, values, locations, mask):

        if not np.any(mask):
            return values # TODO: COPY?

        return_values = values.copy().ravel()
        bound_values = ref_values.copy().ravel()
        coords = locations.reshape(-1, 2)

        if self._rate is None:
            return values # TODO: COPY?

        if len(coords) != len(return_values):
            raise ValueError(
                "Number of locations and values"
                + f" don't match: {len(coords)} vs {len(return_values)}")

        mask_r = mask.copy().ravel()
        nomask_r = np.logical_not(mask_r)

        points = coords[mask_r]
        xy = coords[nomask_r]

        tree = cKDTree(points)
        near_dists, neighbors = tree.query(xy)
        temp_values = return_values[mask_r][neighbors] * (
                1 + near_dists * self._rate * self.rate_sign)

        # NOTE: No bounds are applied for rate
        mask2 = np.logical_not(self.satisfies(
                     return_values[nomask_r], temp_values))
        # Double indexing copies, we want to modify "return_values"
        temp_values_2 = return_values[nomask_r]
        temp_values_2[mask2] = temp_values[mask2]
        return_values[nomask_r] = temp_values_2

        return_values = return_values.reshape(values.shape)
        return  return_values


# TODO:
class BoundMixin:
    pass


class RateMixin:
    pass




class TopoConstConstraint(Constraint):

    def __init__(
            self,
            value,
            upper_bound=np.inf,
            lower_bound=-np.inf,
            value_type: str = 'min',
            rate=None):

        super().__init__(value_type, rate)

        self._lb = lower_bound
        self._ub = upper_bound

        self._val = value


    @property
    def value(self):
        return self._val




    @property
    def topo_bounds(self):

        return self._lb, self._ub


    def apply(self, ref_values, old_values, locations=None):

        lower_bound, upper_bound = self.topo_bounds

        return_values = old_values.copy()

        mask = ((ref_values > lower_bound) &
                (ref_values < upper_bound) &
                (np.logical_not(self.satisfies(return_values, self.value))))
        return_values[mask] = self.value

        return_values = self._apply_rate(ref_values, return_values, locations, mask)

        return return_values



class TopoFuncConstraint(Constraint):

    def __init__(
            self,
            function=lambda i: i / 2.0,
            upper_bound=np.inf,
            lower_bound=-np.inf,
            value_type: str = 'min',
            rate=None):

        super().__init__(value_type, rate)

        self._lb = lower_bound
        self._ub = upper_bound

        self._func = lambda i: i / 2.0
        if callable(function):
            self._func = function


    @property
    def topo_bounds(self):

        return self._lb, self._ub


    def apply(self, ref_values, old_values, locations=None):

        lower_bound, upper_bound = self.topo_bounds

        return_values = old_values.copy()
        temp_values = self._func(ref_values)

        mask = ((ref_values > lower_bound) &
                (ref_values < upper_bound) &
                (np.logical_not(self.satisfies(return_values, temp_values))))
        return_values[mask] = temp_values[mask]

        return_values = self._apply_rate(ref_values, return_values, locations, mask)

        return return_values


class CourantNumConstraint(Constraint):
    '''Class for defining mesh size constraint based on Courant number

    Methods
    -------
    apply
        Calculate and return he new size function at input reference points
    '''

    def __init__(
            self,
            value: float,
            timestep: float = 150,
            wave_amplitude: float = 2.0,
            value_type: str = 'max',
            ):
        '''Constaint for enforcing bound on Courant number

        Parameters
        ----------
        value : float
            The value of Courant number to enforce
        timestep : float, default=150
            The timestep (in seconds) used to calculate Courant number
        wave_amplitude : float, default=2.0
            Amplitude of wave for linear wave theory approximation
        value_type : {'min', 'max'}, default='min'
            Indicate whether to enforce the input value as min or max of Courant #
        '''

        super().__init__(value_type, rate=None)

        self._value = value
        self._dt = timestep
        self._nu = wave_amplitude


    def apply(
        self,
        ref_values,
        old_values,
        *args,
        **kwargs
        ):
        '''Calculate the new values of size function based on input reference

        Parameters
        ----------
        ref_values : array of floats
            Depth values to be used for Courant number approximations
        old_values : array of floats
            Values of mesh size function before applying the constraint
        *args : list
            List of arguments not handled by this apply method (
            used in other constraints)
        **kwargs : dict
            Dictionary of arguments not handled by this apply method (
            used in other constraints)

        Returns
        -------
        array of floats
            New values of size function after application of the constraint
        '''

        if ref_values.shape != old_values.shape:
            raise ValueError("Shapes of depths and sizes arrays don't match")

        return_values = old_values.copy()

        u_mag = utils.estimate_particle_velocity_from_depth(
            ref_values, self._nu
        )
        depth_mask = utils.can_velocity_be_approximated_by_linear_wave_theory(
            ref_values, self._nu
        )
        char_vel = u_mag + np.sqrt(constants.g * np.abs(ref_values))
        # For overland where h < nu the characteristic velocity is 2 * sqrt(g*h)
        char_vel[~depth_mask] = 2 * u_mag[~depth_mask]

        temp_values = utils.get_element_size_courant(
            char_vel, self._dt, self._value
        )
        old_C_apprx = utils.approximate_courant_number_for_depth(
            ref_values, self._dt, old_values, self._nu
        )

        # NOTE: Condition is evaluated on Courant # NOT the element size
        value_mask = np.logical_not(self.satisfies(old_C_apprx, self._value))
        return_values[value_mask] = temp_values[value_mask]

        return return_values
